from abc import ABC, abstractmethod
from io import StringIO
from typing import Generic, TypeVar, overload

T = TypeVar("T")


class Builder(ABC, Generic[T]):
    """Base class for all GraphQL builders.

    A builder is a class that converts an object that represents some GraphQL construct into actual GraphQL code.

    Classes that inherit from this class must implement the `insert` method which takes the source object (of type `T`) and a `StringIO` buffer and inserts the GraphQL code into the buffer.
    """

    def build(self, source: T, /) -> str:
        """Convert a source object into a GraphQL string.

        Args:
            source: The source object to convert.

        Returns:
            The source object as a string of GraphQL code.
        """
        result = StringIO()
        self.insert(source, result)
        return result.getvalue()

    @abstractmethod
    def insert(self, source: T, buffer: StringIO, /) -> None:
        """Insert the GraphQL code for the source object into the result buffer.

        Args:
            source: The source object to convert.
            buffer: The buffer to insert the GraphQL code into.
        """

    @overload
    def __call__(self, source: T, /) -> str: ...

    @overload
    def __call__(self, source: T, buffer: StringIO, /) -> None: ...

    def __call__(self, source: T, buffer: StringIO | None = None) -> str | None:
        if buffer:
            return self.insert(source, buffer)
        return self.build(source)



from typing import Generic, TypeVar

from pydantic import BaseModel, Field

__all__ = ["PageInfo", "Edge", "Connection"]

_T = TypeVar("_T", bound=BaseModel)


class PageInfo(BaseModel):
    """Information about a single page of a paginated connection."""

    has_next_page: bool = Field(alias="hasNextPage")
    """Whether there is a next page."""

    has_previous_page: bool = Field(alias="hasPreviousPage")
    """Whether there is a previous page."""

    start_cursor: str = Field(alias="startCursor")
    """The cursor of the first item in the page."""

    end_cursor: str = Field(alias="endCursor")
    """The cursor of the last item in the page."""


class Edge(BaseModel, Generic[_T]):
    """An edge type for pagination.

    An edge type is used to represent a single item within a paginated connection, and contains the node itself as well as the cursor for that node.
    """

    node: _T
    """The object itself."""

    cursor: str
    """A cursor pointing to the object within the paginated connection."""


class Connection(BaseModel, Generic[_T]):
    """A connection type for pagination.

    This class is used to represent a connection type for pagination in a GraphQL API.
    """

    edges: list[Edge[_T]]
    """A list of edges in the connection."""

    page_info: PageInfo = Field(alias="pageInfo")



from io import StringIO
from typing import Mapping, override

from ..values import GqlValue
from .builder import Builder
from .value_builder import ValueBuilder


class ArgsBuilder(Builder[Mapping[str, GqlValue]]):
    """A GraphQL arguments builder.

    This class is used to convert a mapping of argument names to `GqlValue` objects into a string containing the GraphQL arguments code.
    """

    @override
    def insert(self, args: Mapping[str, GqlValue], buffer: StringIO) -> None:
        """Convert a mapping of argument names to `GqlValue` objects into a GraphQL arguments string.

        Args:
            args: The arguments to convert.

        Returns:
            The arguments as a string.
        """
        buffer.write("(")
        value_builder = ValueBuilder()
        buffer.write(", ".join(f"{k}: {value_builder(v)}" for k, v in args.items()))
        buffer.write(")")



from __future__ import annotations

from typing import Any, Iterable

from .gql_field import GqlField
from .var import Var


class Operation:
    """A GraphQL operation object, such as a query or mutation, which is a collection of fields to be queried from a GraphQL API.

    Args:
        op_type: The type of operation, such as "query" or "mutation".
        op_name: The name of the operation. This has no meaning in the GraphQL operation itself; it is just a label for you to identify the operation.
        fields: The fields to include in the operation. These can be created manually or using the `GqlField.from_model` constructor.
        variables: The variables to include in the operation.
    """

    def __init__(
        self,
        op_type: str,
        op_name: str,
        *fields: GqlField,
        variables: Iterable[Var[Any]] = (),
    ) -> None:
        self.type = op_type
        self.name = op_name
        self.fields = fields
        self.variables = tuple(variables)

    def __format__(self, format_spec: str) -> str:
        from .builders.operation_builder import OperationBuilder

        indent: bool | int | str
        if format_spec in ("", "indent"):
            indent = True
        elif format_spec == "noindent":
            indent = False
        elif format_spec.isdigit():
            indent = int(format_spec)
        elif format_spec.isspace():
            indent = format_spec
        else:
            raise ValueError(
                f"Invalid format specifier: {format_spec!r}. Must be one of '', 'indent', 'noindent', whitespace, or a positive integer."
            )
        return OperationBuilder(indent=indent).build(self)

    def __str__(self) -> str:
        return self.__format__("")

    def __repr__(self) -> str:
        return f"<{self.type} {self:noindent}>"



from io import StringIO
from typing import Any, Sequence, override

from ..var import Var
from .builder import Builder
from .type_builder import TypeBuilder


class VarsBuilder(Builder[Sequence[Var[Any]]]):
    @override
    def insert(self, variables: Sequence[Var[Any]], buffer: StringIO) -> None:
        """Insert variables into the resulting string buffer."""
        if not variables:
            return
        buffer.write("(")
        type_builder = TypeBuilder()
        buffer.write(
            ", ".join(f"${v.name}: {type_builder.build(v)}" for v in variables)
        )
        buffer.write(")")



from __future__ import annotations

from functools import cached_property
from types import UnionType
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Self,
    TypeVar,
    Union,
    _SpecialForm,
    cast,
    get_args,
    get_origin,
    overload,
)

if TYPE_CHECKING:
    from .base_vars import BaseVars

TypeAnnotation = Union[type, _SpecialForm, UnionType]
NOTSET: Any = object()
T = TypeVar("T")


class Var(Generic[T]):
    """A variable definition for a GraphQL query.

    Args:
        name: The name of the variable. If used as a class attribute on a `BaseVars` subclass, the name will be extracted from the attribute name if not provided.
        default: The default value of the variable. If the type annotation allows `None`, the default value will be `None` if not provided. Otherwise, the variable will be considered required.
        type: The python type of the variable. If used as a class attribute on a `BaseVars` subclass, the type will be extracted from the type annotation if not provided. It can also be extracted from the type annotation used when creating the `Var` instance, e.g. `Var[str]("name")`.
        type_name: The name of the type in the GraphQL schema. If provided, it may not contain any `!` or `[]` as these will be added automatically based on type annotations. If not provided, the type name will be determined automatically.
    """

    def __init__(
        self,
        name: str | None = None,
        default: T = NOTSET,
        type: type[T] | None = None,
        type_name: str | None = None,
    ):
        self._default = default
        self._type = type
        self._name = name
        self.type_name = type_name

    @property
    def name(self) -> str:
        """The name of the variable."""
        assert self._name is not None, "Var name is not set"
        return self._name

    @cached_property
    def var_type(self) -> type[T]:
        """The python type of the variable."""
        if self._type:
            return self._type
        if hasattr(self, "__orig_class__"):
            args = get_args(getattr(self, "__orig_class__"))
            if len(args) == 1:
                return args[0]
        raise ValueError("Var type is not set")

    @cached_property
    def required(self) -> bool:
        """Whether the variable is required.

        A variable is required if (1) no default value is set *and* (2) the type annotation does not allow None.
        """
        if self._default is not NOTSET:
            return False
        return is_required(self.var_type)

    @property
    def default(self) -> T:
        """The default value of the variable.

        If the variable is required, accessing this property will raise a `ValueError`.
        """
        if not self.required and self._default is NOTSET:
            return cast(T, None)
        if self.required:
            raise ValueError("No default value set")
        return self._default

    def set_default_name(self, name: str) -> None:
        """Set the name of the variable if it is not set."""
        if self._name is None:
            self._name = name

    def set_default_type(self, var_type: type[T]) -> None:
        """Set the type of the variable if it is not set."""
        if self._type is None:
            self._type = var_type
            self.__dict__.pop("var_type", None)
            self.__dict__.pop("required", None)

    @overload
    def __get__(self, instance: None, owner: type) -> Self: ...
    @overload
    def __get__(self, instance: BaseVars, owner: type) -> T: ...
    def __get__(self, instance: BaseVars | None, owner: type) -> T | Self:
        if instance is None:
            return self
        return instance.__values__[self.name]

    def __set__(self, instance: BaseVars, value: T) -> None:
        instance.__values__[self.name] = value

    def __repr__(self) -> str:
        return f"<Var {self._name or '...'}: {getattr(self._type, '__qualname__', self._type or '...')} {'required' if self.required else f'= {self.default}'}>"


def is_required(t: TypeAnnotation) -> bool:
    """Check if a type annotation is required."""
    return get_origin(t) not in (Union, UnionType) or all(
        arg is not type(None) for arg in get_args(t)
    )



from typing import Self


class Indentation:
    def __init__(self, indentation: str | None = "  ", level: int = 0) -> None:
        self._indentation = indentation
        self._level = level

    def __str__(self) -> str:
        return (
            "\n" + self._indentation * self._level
            if self._indentation is not None
            else ""
        )

    def __add__(self, other: int) -> Self:
        return type(self)(self._indentation, self._level + other)



from dataclasses import dataclass, field
from types import UnionType
from typing import Iterable, Mapping, Self, Sequence, Union, get_args, get_origin

from pydantic import BaseModel
from pydantic.fields import FieldInfo

from .values import GqlValue


@dataclass
class GqlField:
    """A GraphQL field.

    This class is used to represent a field in a GraphQL query. Fields can have arguments and subfields.

    Args:
        name: The name of the field.
        args: A mapping of argument names to argument values.
        fields: A sequence of subfields.
    """

    name: str
    """The name of the field."""

    args: Mapping[str, GqlValue] = field(default_factory=dict)
    """A mapping of argument names to argument values for the field."""

    fields: Sequence[Self] = ()
    """A sequence of subfields for the field."""

    @classmethod
    def from_model(
        cls,
        model: type[BaseModel],
        name: str | None = None,
        args: Mapping[str, GqlValue] = {},
    ) -> Self:
        """Create a `GqlField` from a Pydantic model.

        This method will create a `GqlField` with the same name as the model, and with fields corresponding to the model's fields.

        Args:
            model: The Pydantic model to create a field from.
            name: The name of the field. If not provided, the name of the model will be used.
            args: A mapping of argument names to argument values for the field.

        Returns:
            A `GqlField` representing the model.
        """
        return cls(name or model.__name__, fields=cls.fields_of_model(model), args=args)

    @classmethod
    def fields_of_model(cls, model: type[BaseModel]) -> Sequence[Self]:
        """Get the GraphQL fields corresponding to a Pydantic model."""
        return tuple(
            cls.from_pydantic_field(name, field)
            for name, field in model.model_fields.items()
        )

    @classmethod
    def from_pydantic_field(cls, name: str, field: FieldInfo) -> Self:
        """Create a `GqlField` from a Pydantic field."""
        submodel = _model_of(field)
        return cls(
            name=(
                field.validation_alias
                if isinstance(field.validation_alias, str)
                else name
            ),
            fields=cls.fields_of_model(submodel) if submodel else (),
        )


def _model_of(field: FieldInfo) -> type[BaseModel] | None:
    """Get the model class of a Pydantic field.

    This also supports fields whose type is a collection of nested models.

    For example, for a field of type `NestedModel`, `list[NestedModel]`, etc, this function will return `NestedModel`; but for a field of type `str`, `int`, Iterable[bool], etc, it will return None.
    """
    if field.annotation is None:
        return None
    if isinstance(field.annotation, type) and issubclass(field.annotation, BaseModel):
        return field.annotation
    generic_type = get_origin(field.annotation)
    generic_args = get_args(field.annotation)
    return _check_iterable(generic_type, generic_args) or _check_optional(
        generic_type, generic_args
    )


def _check_iterable(
    generic_type: type | None, generic_args: Sequence[type]
) -> type[BaseModel] | None:
    """Check if a type is an iterable of models."""
    if (
        generic_type
        and issubclass(generic_type, Iterable)
        and len(generic_args) == 1
        and issubclass(generic_args[0], BaseModel)
        and generic_args[0]
    ):
        return generic_args[0]
    return None


def _check_optional(
    generic_type: type | None, generic_args: Sequence[type]
) -> type[BaseModel] | None:
    """Check if a type is an optional model."""
    if (
        generic_type in (Union, UnionType)
        and len(generic_args) == 2
        and type(None) in generic_args
        and any(issubclass(arg, BaseModel) for arg in generic_args)
    ):
        return next(arg for arg in generic_args if issubclass(arg, BaseModel))
    return None



from io import StringIO
from typing import override

from ..operation import Operation
from .builder import Builder
from .fields_builder import FieldsBuilder
from .indentation import Indentation
from .vars_builder import VarsBuilder


class OperationBuilder(Builder[Operation]):
    """A GraphQL operation builder.

    This class is used to convert an `Operation` object into a string containing the GraphQL code.

    Args:
        indent: The indentation to use when formatting the output. If `False` then no indentation is used and the query is returned as a single line. If `True` (the default) then the default indentation is used (two spaces). If an integer then that many spaces are used for indentation. If a string (must be whitespace) then that string is used for indentation.
    """

    DEFAULT_INDENTATION = "  "

    def __init__(self, indent: int | str | bool = True) -> None:
        indentation: str | None
        if isinstance(indent, bool):
            indentation = self.DEFAULT_INDENTATION if indent else None
        elif isinstance(indent, str):
            if not indent.isspace():
                raise ValueError("indent must be whitespace if it is a string.")
            indentation = indent
        else:
            indentation = " " * indent
        self._indentation = Indentation(indentation, 0)

    @override
    def insert(self, operation: Operation, buffer: StringIO) -> None:
        """Insert the operation as a string of GraphQL code into the buffer.

        Args:
            operation: The operation to convert to a string and insert into the buffer.
            buffer: The buffer to write the operation to.
        """
        buffer.write(f"{operation.type} {operation.name}")
        VarsBuilder().insert(operation.variables, buffer)
        buffer.write(" {")
        FieldsBuilder(self._indentation + 1).insert(operation.fields, buffer)
        buffer.write(str(self._indentation))
        buffer.write("}")



from __future__ import annotations

from typing import Any, Iterable, Mapping, Self

from pydantic import BaseModel

from .gql_field import GqlField
from .operation import Operation
from .values import GqlValue
from .var import Var


class Query(Operation):
    """A GraphQL query object, which is a collection of fields to be queried from a GraphQL API.

    # Usage

    To create simple query with one top-level field such as `books`, first create a Pydantic model for the `Book` type, then create a `Query` object with the `Query.from_model` constructor.
    ```python
    class Book(BaseModel):
        title: str
        author: str

    query = Query.from_model(Book, "books")
    print(query)
    ```
    This will output:
    ```
    query Book {
      books {
        title
        author
      }
    }
    ```

    To create a query with multiple top-level fields such as `books` and `authors`, use the `Query` constructor directly.
    ```python
    class Author(BaseModel):
        name: str

    query = Query(
        "BooksAndAuthors",
        GqlField.from_model(Book, "books"),
        GqlField.from_model(Author, "authors"),
    )
    print(query)
    ```
    This will output:
    ```
    query BooksAndAuthors {
      books {
        title,
        author,
      },
      authors {
        name,
      },
    }
    ```

    # Formatting

    To generate the query string to send to the API, use `str(query)`. This will format the query with default indentation of 2 spaces.

    To customize the indentation, use `format` function or f-strings with the desired indentation level or a custom indentation string. The following are all valid:
    ```python
    f"default indentation of 2 spaces: {query}"
    f"same as above:                   {query:indent}"
    f"no indentation; all on one line: {query:noindent}"
    f"indentation of 4 spaces:         {query:4}"
    f"indentation with tabs:           {query:\t}"
    ```

    Args:
        name: The name of the query. This has no meaning in the GraphQL query itself; it is just a label for you to identify the query.
        fields: The fields to include in the query. These can be created manually or using the `GqlField.from_model` constructor.
        variables: The variables to include in the query.
    """

    def __init__(
        self, name: str, *fields: GqlField, variables: Iterable[Var[Any]] = ()
    ) -> None:
        super().__init__("query", name, *fields, variables=variables)

    @classmethod
    def from_model(
        cls,
        model: type[BaseModel],
        field_name: str | None = None,
        query_name: str | None = None,
        variables: Iterable[Var[Any]] = (),
        args: Mapping[str, GqlValue] = {},
    ) -> Self:
        """Create a query with a single top-level field whose subfields are defined by a Pydantic model.

        # Usage

        To create a query with a single top-level field such as `books`, first create a Pydantic model for the `Book` type, then use the `Query.from_model` constructor.

        ```python
        class Book(BaseModel):
            title: str
            author: str

        query = Query.from_model(Book, "books")
        print(query)
        ```
        This will output:
        ```
        query Book {
          books {
            title,
            author,
          },
        }
        ```

        To include variables and arguments in the query, first create a subclass of `BaseVars` with the variables you want to define, then pass the variables and arguments to the constructor.

        ```python
        class Vars(BaseVars):
            year: Var[int]
            genre: Var[str] = Var(default="fiction")

        query = Query.from_model(Book, "books", variables=Vars(year=2021), args={"genre": Vars.genre, "year": Vars.year, "limit": 10})
        print(query)
        ```
        This will output:
        ```
        query Book($year: Int!, $genre: String = "fiction") {
          books(genre: $genre, year: $year, limit: 10) {
            title,
            author,
          },
        }
        ```

        Args:
            model: The Pydantic model to use for the fields.
            field_name: The name of the top-level field. If not provided, the name of the model is used.
            query_name: The name of the query. If not provided, the name of the model is used.
            variables: The variables to include in the query.
            args: The arguments to pass to the top-level field.
        """
        return cls(
            query_name or model.__name__,
            GqlField.from_model(model, field_name, args),
            variables=variables,
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Iterable

from .var import Var


@dataclass
class Expr:
    """An explicit expression for an argument in a GraphQL query.

    Args:
        value: The string representation of the expression in GraphQL.
    """

    value: str

    def __str__(self) -> str:
        return self.value


GqlValue = str | int | float | bool | None | Iterable[Any] | Var[Any] | Expr



from io import StringIO
from types import UnionType
from typing import Any, Iterable, Mapping, Union, cast, get_args, get_origin, override

from ..var import TypeAnnotation, Var, is_required
from .builder import Builder


class TypeBuilder(Builder[Var[Any]]):
    @override
    def insert(self, var: Var[Any], buffer: StringIO) -> None:
        """Insert a type into the resulting string buffer."""
        buffer.write(self._type_of(var.var_type, var.required, var.type_name))

    _basic_types: Mapping[TypeAnnotation, str] = {
        str: "String",
        int: "Int",
        float: "Float",
        bool: "Boolean",
    }

    def _type_of(
        self,
        t: TypeAnnotation,
        required: bool | None,
        type_name: str | None,
    ) -> str:
        """The GraphQL type of the variable as a string."""
        if self._is_optional(t):
            var_type = self._optional_type(t, type_name)
        elif self._is_iterable(t):
            var_type = self._iterable_type(t, type_name)
        elif type_name:
            var_type = type_name
        elif t in self._basic_types:
            var_type = self._basic_types[t]
        elif isinstance(t, type):
            var_type = t.__name__
        else:
            raise ValueError(f"Unsupported type: {t}")
        if required is None:
            required = is_required(t)
        return var_type + ("!" if required else "")

    def _is_optional(self, t: TypeAnnotation) -> bool:
        """Check if a type is an optional."""
        args = get_args(t)
        return (
            get_origin(t) in (Union, UnionType)
            and len(args) == 2
            and type(None) in args
        )

    def _optional_type(self, t: TypeAnnotation, type_name: str | None) -> str:
        """Get the GraphQL type of an Optional type."""
        return self._type_of(
            next(t for t in get_args(t) if t is not type(None)),
            required=False,
            type_name=type_name,
        )

    def _is_iterable(self, t: TypeAnnotation) -> bool:
        """Check if a type is an Iterable of any specific type."""
        try:
            return (
                issubclass(cast(Any, get_origin(t) or t), Iterable)
                and len(get_args(t)) == 1
            )
        except TypeError:
            return False

    def _iterable_type(self, t: TypeAnnotation, type_name: str | None) -> str:
        """Get the GraphQL type of an Iterable."""
        return f"[{self._type_of(get_args(t)[0], required=None, type_name=type_name)}]"



from __future__ import annotations

from abc import ABCMeta
from typing import (
    Any,
    ClassVar,
    Iterator,
    Mapping,
    cast,
    dataclass_transform,
    get_origin,
    get_type_hints,
)

from pydantic import TypeAdapter

from .var import NOTSET, Var


def _is_var(annotation: Any) -> bool:
    """Check if a type annotation (from typing.get_type_hints) is a `Var`."""
    origin = get_origin(annotation)
    return origin is not ClassVar and issubclass(origin or annotation, Var)


@dataclass_transform(eq_default=False, field_specifiers=(Var,), kw_only_default=True)
class BaseVarsMeta(ABCMeta):
    # Can't actually inherit from Iterable. https://discuss.python.org/t/abcmeta-change-isinstancecheck-of-additional-parent-class/19908
    def __iter__(self) -> Iterator[Var[Any]]:
        return iter(cast(type[BaseVars], self).__variables__.values())


class BaseVars(Mapping[str, Any], metaclass=BaseVarsMeta):
    """Base class for classes that define a set of variables for a GraphQL query.

    # Usage

    To define a set of variables, subclass `BaseVars` and define class attributes with type hints of `Var[T]` where `T` is the type of the variable. Optionally, you can assign an instance of `Var` to provide a default value or to override the name and type of the variable.

    If no `Var` instance is assigned, then the variable name will be extracted from the class attribute name, the variable type will be extracted from the type hint, and the default will be `None` if the type hint allows it (e.g. `Optional[T]` or `T | None`). If the type hint does not allow `None`, then the variable will be required and an error will be raised if it is not provided.

    If a `Var` instance is assigned, it may specify a default value for the variable with the keyword argument `default`. If the keyword argument `name` is provided, it will override the variable name. This is useful for variables that have a different name in the GraphQL query than in the Python code, e.g. `local_name: Var[str] = Var(name="remoteName")` will be treated as `remoteName` in the GraphQL query but as an attribute `local_name` in the Python code.

    ```python
    from pydantic_gql import BaseVars, Var

    class MyVars(BaseVars):
        id: Var[int]
        optional: Var[str | None]
        default: Var[str] = Var(default="default value")
        local_name: Var[str] = Var(name="remoteName")
        many: Var[list[str]]
    ```

    The `MyVars` class is itself an iterable of `Var`s. Instances of `MyVars` have values and is a mapping from variable name to a JSON serializable representation of the value. For ma mapping with the original Python values, use `.__values__`.
    """

    __variables__: ClassVar[Mapping[str, Var[Any]]]

    def __init_subclass__(cls) -> None:
        cls.__variables__ = {}
        for key, annotation in get_type_hints(cls).items():
            if not _is_var(annotation):
                continue
            value = getattr(cls, key, NOTSET)
            if isinstance(value, Var):
                value.set_default_name(key)
                value.set_default_type(annotation.__args__[0])
                cls.__variables__[key] = value
            else:
                cls.__variables__[key] = Var(key, value, annotation.__args__[0])
        for key, var in cls.__variables__.items():
            setattr(cls, key, var)

    def __init__(self, *args: object, **kwargs: Any) -> None:
        if args:
            raise TypeError(f"{type(self).__name__} takes no positional arguments")
        self.__values__ = {
            var.name: _get_value(key, kwargs, var)
            for key, var in self.__variables__.items()
        }

    def __iter__(self) -> Iterator[str]:
        return iter(self.__values__)

    def __getitem__(self, name: str, /) -> Any:
        value = self.__values__[name]
        return TypeAdapter(type(value)).dump_python(value, mode="json", by_alias=True)

    def __len__(self) -> int:
        return len(self.__values__)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({', '.join(f'{k}={v!r}' for k, v in self.__values__.items())})"


def _get_value(key: str, kwargs: Mapping[str, Any], var: Var[Any]) -> Any:
    if key in kwargs:
        return kwargs[key]
    try:
        return var.default
    except ValueError:
        raise TypeError(f"missing required variable {key!r}") from None



from __future__ import annotations

from typing import Any, Iterable, Mapping, Self

from pydantic import BaseModel

from .gql_field import GqlField
from .operation import Operation
from .values import GqlValue
from .var import Var


class Mutation(Operation):
    """A GraphQL mutation object, which represents a collection of fields to modify data in a GraphQL API.

    # Usage

    To create a simple mutation with one top-level field such as `add_book`, first create a Pydantic model for the `Book` type, then create a `Mutation` object with the `Mutation.from_model` constructor.
    ```python
    class Book(BaseModel):
        title: str
        author: str

    book = Book(title="The Lord of the Rings", author="J.R.R. Tolkien")

    mutation = Mutation.from_model(Book, "add_book", args=dict(book))
    print(mutation)
    ```
    This will output:
    ```
    mutation AddBook {
      add_book(title: "The Lord of the Rings", author: "J.R.R. Tolkien") {
        title,
        author,
      },
    }
    ```

    To create a mutation with multiple top-level fields such as `add_book` and `remove_book`, use the `Mutation` constructor directly.
    ```python
    mutation = Mutation(
        "ModifyBooks",
        GqlField.from_model(Book, "add_book", args=dict(book)),
        GqlField("remove_book", args={"id": 1}),
    )
    print(mutation)
    ```
    This will output:
    ```
    mutation ModifyBooks {
      add_book(title: "The Lord of the Rings", author: "J.R.R. Tolkien") {
        title,
        author,
      },
      remove_book(id: 1),
    }
    ```

    # Formatting

    To generate the mutation string to send to the API, use `str(mutation)`. This will format the mutation with default indentation of 2 spaces.

    To customize the indentation, use `format` function or f-strings with the desired indentation level or a custom indentation string. The following are all valid:
    ```python
    f"default indentation of 2 spaces: {mutation}"
    f"same as above:                   {mutation:indent}"
    f"no indentation; all on one line: {mutation:noindent}"
    f"indentation of 4 spaces:         {mutation:4}"
    f"indentation with tabs:           {mutation:\t}"
    ```

    Args:
        name: The name of the mutation. This has no meaning in the GraphQL mutation itself; it is just a label for you to identify the mutation.
        fields: The fields to include in the mutation. These can be created manually or using the `GqlField.from_model` constructor.
        variables: The variables to include in the mutation.
    """

    def __init__(
        self, name: str, *fields: GqlField, variables: Iterable[Var[Any]] = ()
    ) -> None:
        super().__init__("mutation", name, *fields, variables=variables)

    @classmethod
    def from_model(
        cls,
        model: type[BaseModel],
        field_name: str | None = None,
        mutation_name: str | None = None,
        variables: Iterable[Var[Any]] = (),
        args: Mapping[str, GqlValue] = {},
    ) -> Self:
        """Create a mutation with a single top-level field whose subfields are defined by a Pydantic model.

        # Usage

        To create a mutation with a one top-level field such as `add_book`, first create a Pydantic model for the `Book` type, then use the `Mutation.from_model` constructor.

        ```python
        class Book(BaseModel):
            title: str
            author: str

        book = Book(title="The Lord of the Rings", author="J.R.R. Tolkien")

        mutation = Mutation.from_model(Book, "add_book", args=dict(book))
        print(mutation)
        ```
        This will output:
        ```
        mutation Book {
          add_book(title: "The Lord of the Rings", author: "J.R.R. Tolkien") {
            title,
            author,
          },
        }
        ```

        To include variables and arguments in the query, first create a subclass of `BaseVars` with the variables you want to define, then pass the variables and arguments to the constructor.

        ```python
        class RemovedBook(BaseModel):
            id: int
            archived: bool

        class Vars(BaseVars):
            id: Var[int | None] = Var(default=None)
            title: Var[str | None] = Var(default=None)

        variables = Vars(title="The Lord of the Rings")
        mutation = Mutation.from_model(
            RemovedBook,
            "remove_book",
            variables=variables,
            args={"id": Vars.id, "title": Vars.title, "archive": True},
        )
        print(mutation)
        ```
        This will output:
        ```
        mutation RemovedBook($id: Int, $title: String) {
          remove_book(id: $id, title: $title, archive: true) {
            id,
            archived,
          },
        }
        ```

        Args:
            model: The Pydantic model to use for the fields.
            field_name: The name of the top-level field. If not provided, the name of the model is used.
            mutation_name: The name of the mutation. If not provided, the name of the model is used.
            variables: The variables to include in the mutation.
            args: The arguments to pass to the top-level field.
        """

        return cls(
            mutation_name or model.__name__,
            GqlField.from_model(model, field_name, args),
            variables=variables,
        )



"""
.. include:: ../README.md
"""

import importlib.metadata as metadata

__version__ = metadata.version(__package__ or __name__)

from .base_vars import BaseVars
from .gql_field import GqlField
from .mutation import Mutation
from .query import Query
from .values import Expr, GqlValue
from .var import Var

__all__ = ("Query", "Mutation", "BaseVars", "Var", "GqlField", "Expr", "GqlValue")



from io import StringIO
from typing import Iterable, override

from ..gql_field import GqlField
from .args_builder import ArgsBuilder
from .builder import Builder
from .indentation import Indentation


class FieldsBuilder(Builder[Iterable[GqlField]]):
    """A GraphQL fields builder.

    This class is used to convert a sequence of `GqlField` objects into a string containing the GraphQL fields code.
    """

    def __init__(self, indentation: Indentation) -> None:
        self._indentation = indentation

    @override
    def insert(self, fields: Iterable[GqlField], buffer: StringIO) -> None:
        """Convert a sequence of `GqlField` objects into a GraphQL fields string.

        Args:
            fields: The fields to convert.

        Returns:
            The fields as a string.
        """
        for field in fields:
            self._insert_field(field, buffer)

    def _insert_field(self, field: GqlField, buffer: StringIO) -> None:
        """Insert one field into the resulting string buffer."""
        buffer.write(f"{self._indentation}{field.name}")
        if field.args:
            ArgsBuilder().insert(field.args, buffer)
        if field.fields:
            buffer.write(" {")
            FieldsBuilder(self._indentation + 1).insert(field.fields, buffer)
            buffer.write(str(self._indentation))
            buffer.write("}")
        buffer.write(",")



from .args_builder import ArgsBuilder
from .builder import Builder
from .fields_builder import FieldsBuilder
from .indentation import Indentation
from .operation_builder import OperationBuilder
from .type_builder import TypeBuilder
from .value_builder import ValueBuilder
from .vars_builder import VarsBuilder

__all__ = (
    "ArgsBuilder",
    "Builder",
    "FieldsBuilder",
    "Indentation",
    "OperationBuilder",
    "TypeBuilder",
    "VarsBuilder",
    "ValueBuilder",
)



import json
from io import StringIO
from typing import Iterable, override

from ..values import GqlValue
from ..var import Var
from .builder import Builder


class ValueBuilder(Builder[GqlValue]):
    """A GraphQL value builder.

    This class is used to convert a `GqlValue` object into a string containing the GraphQL value code.
    """

    @override
    def insert(self, value: GqlValue, buffer: StringIO) -> None:
        """Convert a `GqlValue` object into a GraphQL value string.

        Args:
            value: The value to convert.

        Returns:
            The value as a string.
        """
        if isinstance(value, Var):
            buffer.write(f"${value.name}")
        elif isinstance(value, str):
            buffer.write(json.dumps(value))
        elif isinstance(value, bool):
            buffer.write(str(value).lower())
        elif value is None:
            buffer.write("null")
        elif isinstance(value, Iterable):
            buffer.write(f"[{', '.join(self.build(v) for v in value)}]")
        else:
            buffer.write(str(value))
